﻿using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Text;
using System.Windows.Forms;

namespace MapLcdFont
{
    public partial class MapLcdFont : UserControl {

        /// <summary>
        /// Изображение фона таблицы
        /// </summary>
        private Bitmap _bmp = new Bitmap(1, 1);
        /// <summary>
        /// Спрайт "незакрашенная точка"
        /// </summary>
        private Bitmap _dotfon = new Bitmap(3, 3);
        /// <summary>
        /// Спрайт "закрашенная точка
        /// </summary>
        private Bitmap _dot = new Bitmap(3, 3);
        /// <summary>
        /// расстояния между точками
        /// </summary>
        private int _dx = 5 + 3 * 5;
        private int _dy = 3 + 3 * 8;

        public MapLcdFont()
        {
            InitializeComponent();
            SetStyle(ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint | ControlStyles.OptimizedDoubleBuffer | ControlStyles.ResizeRedraw, true);        
            Draw_dot_and_dotfon();
            DrawFon();
            MapPanel.Invalidate();
        }
        #region Init draw fon, dot, dotfon
        private void Draw_dot_and_dotfon()
        {
            Bitmap a = new Bitmap(2, 2);
            Graphics e = Graphics.FromImage(a);
            // закрашивание фона точки, с определенной контрастностью
            e.Clear(Color.FromArgb(_contr, 0, 0, 0)); // контрастность имитируется за счет создания черного фона с определенной прозрачностью...
            e = Graphics.FromImage(_dotfon);
            e.Clear(BackColor);
            e.DrawImage(a, 0, 0, 3, 3);// ...и наложением его на фон компонента.
            e = Graphics.FromImage(a);
            e.Clear(_dotColor);
            e = Graphics.FromImage(_dot);
            e.Clear(BackColor);
            e.DrawImage(a, 0, 0, 3, 3);
            a.Dispose();
            e.Dispose();
        }
        private void DrawFon()
        {
            _bmp = new Bitmap(18 + 16 * _dx, 16 * _dy + 1); // создание битмапа на 256 знакомест, с учетом колонки под обозначение по оси Y
            Width = _bmp.Width + 18; // Увеличиваем контрол и PictureBox до необходимых размеров
            MapPanel.Size = _bmp.Size;
            Graphics g = Graphics.FromImage(_bmp);
            g.Clear(BackColor);
            g.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;
            Pen p = new Pen(_gridColor);
            g.DrawLine(p, 0, 0, 0, _bmp.Height - 1); // слева вертикальная линия сетки
            g.DrawLine(p, _bmp.Width - 1, 0, _bmp.Width - 1, _bmp.Height - 1); // справа вертикальная линия сетки
            g.DrawLine(p, 0, _bmp.Height - 1, _bmp.Width - 1, _bmp.Height - 1); // снизу горизонтальная линия сетки
            for (byte i = 0; i < 16; i++)
            {
                g.DrawLine(p, _dx * i + 17, 0, _dx * i + 17, _bmp.Height - 1); // вертикальная линия сетки
                g.DrawString("0" + i.ToString("X"), Font, new SolidBrush(ForeColor), 0, ((_dy - Font.Height) >> 1) + i * _dy);// обозначения по вертикали
                if (i != 0) // горизонтальные линии. Верхняя не рисуется, так как она есть у панели с верхними обозначениями
                    g.DrawLine(p, 0, i * _dy, _bmp.Width - 1, i * _dy);
            }
            p.Dispose();
            g.Dispose();
        }
        #endregion Init draw fon, dot, dotfon
        #region events Mouse Click
        /// <summary>
        /// Обработка клика мышью
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void MapPanelMouseClick(object sender, MouseEventArgs e) {
            if ((e.X < 18) || (e.X > (_dx * 16 + 16)) || (e.Y > (_dy * 16 - 1)))
            {
                return; // клик произошел за прелами карты знакомест
            }
            _posX = (byte)((e.X - 17) / _dx);
            _posY = (byte)(e.Y / _dy);
            // небольшая проверка целостности массива и отправка события с координатами и символом LcdChar
            if (_lcd_font.Chars.Length == 256)
                OnSelectedChar(new SelectedCharEventArgs(new Point(_posX, _posY), _lcd_font.Chars[_posX, _posY]));
            MapPanel.Invalidate();
            Focus();
        }
        
        public delegate void SelectedCharEventHandler(object sender, SelectedCharEventArgs e);
        [Category("Action")]
        public event EventHandler<SelectedCharEventArgs> SelectedChar;
        protected virtual void OnSelectedChar(SelectedCharEventArgs e) => SelectedChar?.Invoke(this, e);
        #endregion events Mouse Click
        #region Paint
        /// <summary>
        /// Отрисовка верхней панели с обозначениями оси Х
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void TopPanelPaint(object sender, PaintEventArgs e)
        {
            e.Graphics.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;
            e.Graphics.Clear(BackColor);
            Pen p = new Pen(_gridColor);
            e.Graphics.DrawRectangle(p, 0, 0, 337, 15);// прямоугольник по периметру
            for (byte i = 0; i < 16; i++)
            {  
                e.Graphics.DrawString(i.ToString("X") + "0", Font, new SolidBrush(ForeColor), 20 * i + 19, 1);// обозначения
                e.Graphics.DrawLine(p, 20 * i + 17, 0, 20 * i + 17, 15);// вертикальные линие
            }
            p.Dispose();
        }
        /// <summary>
        /// отрисовка шрифта
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void MapPanelPaint(object sender, PaintEventArgs e)
        {            
            e.Graphics.DrawImageUnscaled(_bmp, 0, 0);// отрисовка фона
            try
            {                
                if (_lcd_font.Chars.Length != 256) return;// неидеальная, но всё же проверка на необходимое количество элементов в массиве
                for (byte x = 0; x < 16; x++) //перебор элементов(символов) шрифта
                    for (byte y = 0; y < 16; y++)
                    {
                        
                        
                        
                        
                        if (_lcd_font.Chars[x, y].Dots.Length != 40) return; // проверка целостности массива точек
                        
                        for (byte xx = 0; xx < 5; xx++)
                            for (byte yy = 0; yy < 8; yy++) //отрисовка точки (или фона точки)
                                e.Graphics.DrawImageUnscaled(_lcd_font.Chars[x, y].Dots[xx, yy] ? _dot : _dotfon, x * _dx + 3 * xx + 20, y * _dy + 3 * yy + 2);
                    }
                // отрисовка прямоугольника вокруг выделенного символа, цвет изменяется в зависимости от свойства "только для чтения"
                e.Graphics.DrawRectangle(new Pen(_lcd_font.Chars[_posX, _posY].User ? _colorFocusEdit : _colorFocus), 17 + _posX * _dx, _posY * _dy, _dx, _dy);
                // если у знакоместа есть дубль, то периметр дубля так же выделяется
                Point a = _lcd_font.Chars[_posX, _posY].DubAddr;
                if (a.X != -1)
                {
                    e.Graphics.DrawRectangle(new Pen(_lcd_font.Chars[a.X, a.Y].User ? _colorFocusEdit : _colorFocus), 17 + a.X * _dx, a.Y * _dy, _dx, _dy);
                }
            }
            catch { }
        }
        #endregion Paint        
        #region propertys
        /// <summary>
        /// Непосредсвенно шрифт. Присваивание сделано напрямую, хотя и ссылочный тип. Это надо учитывать.
        /// </summary>
        private LcdFont _lcd_font = new LcdFont();
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public LcdFont LcdFont
        {
            get => _lcd_font;
            set
            {
                _lcd_font = value;
                MapPanel.Invalidate();
            }
        }
        /// <summary>
        ///  Выделенный символ. Присваивание сделано напрямую, хотя и ссылочный тип. Это надо учитывать.
        /// </summary>
        [DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
        public LcdChar SelChar
        {
            get => _lcd_font.Chars[_posX, _posY];
            set
            {
                _lcd_font.Chars[_posX, _posY] = value;
                MapPanel.Invalidate();
            }
        }
        private Color _dotColor = Color.Black;
        [DefaultValue(typeof(Color), "Black"), Category("LCD"), Description("")]
        public Color DotColor
        {
            get => _dotColor;
            set
            {
                _dotColor = value;
                DrawFon();
                MapPanel.Invalidate();
                TopPanel.Invalidate();
            }
        }
        private Color _gridColor = Color.FromArgb(180, 180, 180);
		[DefaultValue(typeof(Color), "180, 180, 180"), Category("LCD"), Description("")]
		public Color GridColor
        {
            get => _gridColor;
            set
            {
                _gridColor = value;
                DrawFon();
                MapPanel.Invalidate();
                TopPanel.Invalidate();
            }
        }
        private Color _colorFocusEdit = Color.Green;
		[DefaultValue(typeof(Color), "Green"), Category("LCD"), Description("")]
		public Color ColorFocusEdit
        {
            get => _colorFocusEdit;
            set
            {
                _colorFocusEdit = value;
                MapPanel.Invalidate();
            }
        }

        private Color _colorFocus = Color.Red;
		[DefaultValue(typeof(Color), "red"), Category("LCD"), Description("")]
		public Color ColorFocus
        {
            get => _colorFocus;
            set
            {
                _colorFocus = value;
                MapPanel.Invalidate();
            }
        }

        private byte _posX = 0;
		[DefaultValue(typeof(byte), "0"), Category("LCD"), Description("")]
		public byte PosX
        {
            get => _posX;
            set
            {
                _posX = value;
                MapPanel.Invalidate();
            }
        }

        private byte _posY = 0;
		[DefaultValue(typeof(byte), "0"), Category("LCD"), Description("")]
		public byte PosY
        {
            get => _posY;
            set
            {
                _posY = value;
                MapPanel.Invalidate();
            }
        }

        private byte _contr = 50;
        [DefaultValue(typeof(byte), "50"), Category("LCD"), Description("")]
        public byte DotsContrast
        {
            get => _contr;
            set
            {
                _contr = value;
                Draw_dot_and_dotfon();
                MapPanel.Invalidate();
            }
        }
        #endregion
        public void Redraw() => MapPanel.Invalidate();
        private void Back_Fore_ColorChanged(object sender, EventArgs e)
        {            
            Draw_dot_and_dotfon();
            DrawFon();
            MapPanel.Invalidate();
            TopPanel.Invalidate();
        }
    }
}